#lang racket/base
(require racket/include)

(define (command-line-arguments)
  (vector->list (current-command-line-arguments)))
(define (print-extended-identifiers on?) (void))

(define impl (vector-ref (current-command-line-arguments) 2))

(define flush-output-port flush-output)

(include "../ca/setup.ss")
(set! f f) ; prevent inlining

'----------------------------------------

;; Baseline loop performance
(show 'base-loop)
(times
 (let loop ([i N])
   (if (zero? i)
       0
       (loop (sub1 i)))))

;; Baseline for continuation-growing non-tail recursion (note `Q` instead of `N`)
(show 'base-deep)
(times
 (for/fold ([v #f]) ([j (in-range 100)])
   (let loop ([i Q])
     (if (zero? i)
         0
         (add1 (loop (sub1 i)))))))

;; Racket CS time is 12% slower than CS time due to
;;  the extra code generated by `#%app`
(show 'base-arg-call-loop)
(times
 (let ([f f])
   (let loop ([i N])
     (if (zero? i)
         0
         (loop (f (sub1 i)))))))

'----------------------------------------

;; Continuation marks around tail call for a loop
;; Should be similar to 'attachment-consume-pair-set-loop
;;  from "attach.ss", but extra overhead is call to an "update"
;;  function instead of an inline `cons`
(show 'set-loop)
(times
 (let loop ([i N])
   (if (zero? i)
       (continuation-mark-set-first #f 'key)
       (with-continuation-mark
        'key i
        (loop (sub1 i))))))

;; Analogous to 'attachment-set-nontail-easy
;; Racket CS: in large-M configuration, nearly all
;;  time is GC due grouping attachment list
(show 'set-nontail-prim)
(times
 (for/fold ([v #f]) ([j (in-range 10)])
   (let loop ([i Q])
     (if (zero? i)
         (continuation-mark-set-first #f 'key)
         (add1
          (with-continuation-mark
           'key i
           (add1 (loop (sub1 i)))))))))

;; Racket: 1/3 time is GC
;; Racket CS: nearly all time is GC, because
;;  this creates a chain of continuation records
(show 'set-tail-notail)
(times
 (for/fold ([v #f]) ([j (in-range 10)])
   (let loop ([i Q])
     (if (zero? i)
         (continuation-mark-set-first #f 'key)
         ;; Note: traditional Racket uses a kind of `begin0` here
         (with-continuation-mark
          'key i
          (add1 (loop (sub1 i))))))))

;; Racket: 1/2 time is GC
;; Racket CS: nearly all time is GC, because
;;  this creates a chain of continuation records
(show 'set-nontail-tail)
(times
 (for/fold ([v #f]) ([j (in-range 10)])
   (let loop ([i Q])
     (if (zero? i)
         (continuation-mark-set-first #f 'key)
         (add1
          (with-continuation-mark
           'key i
           (loop (sub1 i))))))))

;; Racket CS time is 25% slower than CS time due to
;;  `#%app` plus the `cons` of 'key and `i`
(show 'set-arg-call-loop)
(times
 (let ([f f])
   (let loop ([i N])
     (if (zero? i)
         0
         (loop (with-continuation-mark
                'key i
                (f (sub1 i))))))))
 
;; Racket CS time is 45% slower than CS time...
(show 'set-arg-prim-loop)
(times
 (let ([f f])
   (let loop ([i N])
     (if (zero? i)
         0
         (loop (f (with-continuation-mark
                   'key i
                   (sub1 i))))))))

'----------------------------------------

;; Lookup with key not found
(show 'first-none-loop)
(times
 (let loop ([i L] [a #f])
   (if (zero? i)
       a
       (loop (sub1 i) (continuation-mark-set-first #f 'key)))))

;; Lookup with key found
(show 'first-some-loop)
(times
 (with-continuation-mark
  'key 'val
  (let loop ([i L] [a #f])
    (if (zero? i)
        a
        (loop (sub1 i) (continuation-mark-set-first #f 'key))))))

;; Lookup with key found in distant frame with lots of other
;;  keys on frames in between
(show 'first-deep-loop)
(times
 (with-continuation-mark
  'key 'val
  (let loop ([j 100])
    (cond
      [(zero? j)
       (let loop ([i L] [a #f])
         (if (zero? i)
             a
             (loop (sub1 i) (continuation-mark-set-first #f 'key))))]
      [(odd? j)
       (list  (with-continuation-mark
               'other 'val
               (loop (sub1 j))))]
      [else
       (car (loop (sub1 j)))]))))

;; ----------------------------------------

;; Key not found
(show 'immed-none-loop)
(times
 (let loop ([i N] [a #f])
   (if (zero? i)
       a
       (call-with-immediate-continuation-mark
        'key
        (lambda (a)
          (loop (sub1 i) a))))))

;; Key found
(show 'immed-some-loop)
(times
 (with-continuation-mark
  'key 'val
  (let loop ([i N] [a #f])
    (if (zero? i)
        a
        (call-with-immediate-continuation-mark
         'key
         (lambda (a)
           (loop (sub1 i) a)))))))
