#lang racket/base
(require racket/format)

;; The `parse-times` function reads the output of running benchamrks
;; and extracts run times, stdandard deviation, etc.

(provide parse-times)

(define (parse-times in report
                     #:drop-runs [drop-runs 0]
                     #:accum [accum cons]
                     #:base [base '()]
                     #:rel-stdev? [rel-stdev? #f]
                     #:gc? [gc? #f])
  (let loop ([group #f] [impl #f] [times null])
    (define l (read-line in))
    (cond
      [(eof-object? l)
       (if (or group impl)
           (accum (summary report group impl times drop-runs rel-stdev?)
                  base)
           base)]
      [else
       (cond
         [(regexp-match #rx"== ([a-zA-Z0-9-]+) ([a-zA-Z0-9-]+) ([0-9]+) ==" l)
          => (lambda (m)
               (define first (when group
                               (summary report group impl times drop-runs rel-stdev?)))
               (define rest (loop (cadr m) (caddr m) '()))
               (if group
                   (accum first rest)
                   rest))]
         [(and gc?
               (regexp-match #rx"elapsed cpu time, including ([0-9.]+)s collecting" l))
          => (lambda (m)
               (loop group impl (cons (inexact->exact (floor (* 1000 (string->number (cadr m)))))
                                      times)))]
         [(regexp-match #rx"cpu (?:time|msec): ([0-9]+)" l)
          => (lambda (m)
               (loop group impl (cons (string->number (cadr m)) times)))]
         [(or (regexp-match #rx"([0-9.]+)(?:s| secs) (?:CPU time|elapsed cpu time|cpu time)" l)
              (regexp-match #rx"elapsed ([0-9.]+)s" l))
          => (lambda (m)
               (loop group impl (cons (inexact->exact (floor (* 1000 (string->number (cadr m)))))
                                      times)))]
         [(regexp-match #rx"([0-9.]+)u ([0-9.]+)s" l) ;  [0-9]+:[0-9.]+ [0-9.]+%" l)
          => (lambda (m)
               (loop group impl (cons (inexact->exact (floor (* 1000 (+ (string->number (cadr m))
                                                                        (string->number (caddr m))))))
                                      times)))]
         [else (loop group impl times)])])))

(define (summary report group impl times drop-runs rel-stdev?)
  (when (null? times)
    (error 'summary "missing data for ~s ~s" group impl))
  (let* ([times (reverse times)]
         [times (if ((length times) . > . drop-runs)
                    (list-tail times drop-runs)
                    '())])
    (define median (if (null? times)
                       "???"
                       (list-ref (sort times <) (quotient (length times) 2))))
    (define avg (if (null? times)
                    "???"
                    (quotient (apply + times) (length times))))
    (define dev (cond
                  [(null? times) "???"]
                  [(null? (cdr times)) (if rel-stdev?
                                           "0%"
                                           "±0")]
                  [else
                   (let ([stdev (sqrt (/ (for/sum ([t (in-list times)])
                                           (expt (- t avg) 2))
                                         (sub1 (length times))))])
                     (cond
                       [rel-stdev?
                        (string-append
                         (~r #:precision '(= 2)
                             (* 100
                                (/ stdev
                                   avg)))
                         "%")]
                       [else
                        (format "±~a" (inexact->exact (round stdev)))]))]))
    (report group impl (number->string median) (number->string avg) dev)))

(module+ main
  (require racket/cmdline)

  (define NAME-WIDTH 25)
  (define TIME-WIDTH 8)
  (define RSD-WIDTH 10)

  (define line-format "~a~a ~a ~a\n")

  (define drop-runs 0)
  (define gc? #f)
  (define sort? #f)

  (define lines null)

  (command-line
   #:once-each
   [("--drop") n "Drop first <n> runs"
               (set! drop-runs (string->number n))]
   [("--width") w "Name width as <w>"
                (set! NAME-WIDTH (string->number w))]
   [("--gc") "Collect GC times"
             (set! gc? #t)]
   [("--sort") "Sort by average"
               (set! sort? #t)]
   #:args ()
   (void))

  (define (report group impl median avg dev)
    (printf line-format
            (pad (format "~a ~a:" group impl) NAME-WIDTH)
            (pad median TIME-WIDTH)
            (pad avg TIME-WIDTH)
            (pad dev RSD-WIDTH)))

  (define (save-or-report group impl median avg dev)
    (cond
      [sort?
       (set! lines (cons (list group impl median avg dev)
                         lines))]
      [else (report group impl median avg dev)]))

  (define (pad s n)
    (let ([s (format "~a" s)])
      (string-append (make-string (max 0 (- n (string-length s))) #\space) s)))

  (printf line-format
          (pad "" NAME-WIDTH)
          (pad "median" TIME-WIDTH)
          (pad "mean" TIME-WIDTH)
          (pad "stdev" RSD-WIDTH))

  (parse-times (current-input-port)
               #:drop-runs drop-runs
               save-or-report
               #:accum void
               #:gc? gc?)

  (when sort?
    (for ([l (in-list (sort lines <
                            #:key (lambda (p)
                                    (string->number (list-ref p 3)))))])
      (apply report l))))
