#lang racket/base
(require scribble/base
         (only-in scribble/core
                  content->string
                  style
                  element
                  color-property)
         racket/format
         "summary.rkt")

(provide ctak-table
         triple-table
         extra-triple-table
         attach-table
         mark-table
         pycket-mark-table
         opt-mark-table
         scheme-table
         ctc-table
         opt-ctc-table
         app-table
         opt-app-table)

(define (call-with-input-file/maybe file proc)
  (cond
    [(file-exists? file)
     (call-with-input-file* file proc)]
    [else
     (proc (open-input-bytes #""))]))

(define (ctak-table)
  (make-table "bmtimes/ctak.txt"))

(define (triple-table)
  (make-table "bmtimes/triple.txt"
              #:omits '("chez-cps" "chez-current" "chez-attach")
              #:add-group
              (lambda (i g)
                (string-append
                 i
                 " "
                 (group->name g)))))

(define (extra-triple-table)
  (make-table "bmtimes/chez-triple.txt"
              #:omits '("chez-raw-attach")
              #:exe->name (lambda (s)
                            (case s
                              [("chez-current") "unmodified"]
                              [("chez-raw-attach") "just attach"]
                              [("chez-attach") "attach"]
                              [("chez") "CS"]
                              [else s]))
              #:row<? (lambda (r1 r2)
                        (define (group g) (case g
                                            [("scheme-oleg") 0]
                                            [("scheme-monadic") 1]
                                            [else 2]))
                        (define (impl i) (case i
                                           [("chez-current") 0]
                                           [("chez-raw-attach") 1]
                                           [("chez-attach") 2]
                                           [("chez") 3]
                                           [else 4]))
                        (or (< (group (car r1)) (group (car r2)))
                            (and
                             (equal? (car r1) (car r2))
                             (< (impl (cadr r1)) (impl (cadr r2))))))
              #:add-group
              (lambda (i g)
                (string-append
                 i
                 " "
                 (group->name g)))))

(define (attach-table)
  (factor-table "bmtimes/ca.txt"
                #:base "imitation"
                #:next "builtin"
                #:base-label "imitate"
                #:next-label "builtin"
                #:gray-factors '("base-loop"
                                 "base-callcc-loop"
                                 "base-deep"
                                 "base-callcc-deep")))

(define (mark-table)
  (factor-table "bmtimes/cm.txt"
                #:base "racket"
                #:next "racketcs"
                #:base-label "Racket"
                #:next-label "CS"))

(define (pycket-mark-table)
  (factor-table "bmtimes/cm.txt"
                #:base "pycket"
                #:next "racketcs"
                #:base-label "Pycket"
                #:next-label "CS"))

(define opts '("racketcs-no-opt1cc" "racketcs-no-caopt" "racketcs-no-caprimopt"))
(define opt-next-labels '("no" "no" "no"))
(define opt-time-labels '("1cc" "opt" "prim"))

(define (opt-mark-table)
  (factor1-table (call-with-input-file/maybe
                  "bmtimes/cm.txt"
                  (lambda (in)
                    (parse-times in list #:rel-stdev? #t)))
                 #:label "benchmark"
                 #:base "racketcs"
                 #:nexts opts
                 #:base-label "Racket"
                 #:next-labels opt-next-labels
                 #:time-labels opt-time-labels
                 #:skip-bms '("base-loop"
                              "base-callcc-loop"
                              "base-arg-call-loop")
                 #:relstdev-label "stdev"
                 #:sep (hspace 1)
                 #:ratio-prefix (hspace 1)))

(define (factor-table file
                      #:base base-name
                      #:next next-name
                      #:base-label base-label
                      #:next-label next-label
                      #:gray-factors [gray-factors '()])
  (define rows
    (if (list? file)
        file
        (call-with-input-file/maybe
         file
         (lambda (in)
           (parse-times in list #:rel-stdev? #t)))))

  (define (->n s) (string->number s))
  (define (%->n s) (/ (string->number (substring s 0 (sub1 (string-length s)))) 100.0))
  (define (fmt n) (factor (~r #:precision '(= 1) n)))
  (define (fmt% n) (format "~a%" (inexact->exact (floor (* n 100)))))

  (define ratios
    (for/list ([row (in-list rows)]
               #:when (equal? next-name (cadr row)))
      (define i-row (or (for/first ([i-row (in-list rows)]
                                    #:when (and (equal? base-name (cadr i-row))
                                                (equal? (car row) (car i-row))))
                          i-row)
                        (begin
                          (log-error "warning: missing ~s ~s" base-name (car row))
                          row)))
      (define avg (->n (list-ref row 3)))
      (define var (%->n (list-ref row 4)))
      (define i-avg (->n (list-ref i-row 3)))
      (define i-var (%->n (list-ref i-row 4)))

      (define ratio (if (zero? avg) avg (/ i-avg avg)))

      (define gray? (member (car row) gray-factors))
      (define (maybe-as-gray c)
        (if gray?
            (element (style #f (list (color-property "gray"))) c)
            c))

      (list (tt (regexp-replace #rx"attachment-" (car row) ""))
            (list (hspace 2) (number->string avg) " ms")
            (list (hspace 2) (maybe-as-gray (fmt ratio)))
            (list (hspace 2)
                  (maybe-as-gray
                   (fmt (if (zero? avg)
                            0.0
                            (/ (* (- 1 i-var) i-avg)
                               (* (+ 1 var) avg))))))
            (maybe-as-gray (list "-"))
            (list (maybe-as-gray
                   (fmt (if (zero? avg)
                            0.0
                            (/ (* (+ 1 i-var) i-avg)
                               (* (- 1 var) avg)))))))))

  (tabular
   #:style 'block
   #:column-properties '(right)
   (list*
    (map header (list "" next-label (list (hspace 2) base-label) (list (hspace 2) "speedup") 'cont 'cont))
    (map header (list "benchmark" "time" "time" (list (hspace 2) "range") 'cont 'cont))
    ratios)))

(define (make-table bmtimes-file
                    #:omits [omits '()]
                    #:only [only #f]
                    #:drop-runs [drop-runs 0]
                    #:add-group [add-group (lambda (i g) i)]
                    #:exe->name [exe->name exe->name]
                    #:rel-stdev? [rel-stdev? #f]
                    #:row<? [row<? (lambda (row1 row2) (< (string->number (caddr row1))
                                                          (string->number (caddr row2))))])
  (define rows
    (filter
     (lambda (row) (and (not (member (car row) omits))
                        (not (member (cadr row) omits))
                        (or (not only)
                            (member (car row) only)
                            (member (cadr row) only))))
     (sort
      (call-with-input-file/maybe
       bmtimes-file
       (lambda (in)
         (parse-times in list
                      #:drop-runs drop-runs
                      #:rel-stdev? rel-stdev?)))
      row<?)))

  (tabular
   #:style 'block
   #:column-properties '(right)
   #:sep (hspace 2)
   (cons
    (map header
         (list ""
               #; "median"
               "average"
               (if rel-stdev? "relstdev" "stdev")))
    (for/list ([row (in-list rows)])
      (list (add-group (exe->name (cadr row))
                       (car row))
            #; (string-append (list-ref row 2) " ms")
            (string-append (list-ref row 3) " ms")
            (string-append (list-ref row 4) " ms"))))))

(define (exe->name s)
  (case s
    [("racketcs") "Racket CS"]
    [("racketcs-imitate") "Racket CS imitate"]
    [("chez") "Chez Scheme"]
    [("ghc") "GHC"]
    [("multicore-ocaml") "Multicore OCaml"]
    [("koka") "Koka on Node.js"]
    [("chicken") "CHICKEN"]
    [else (string-titlecase s)]))

(define (group->name s)
  (case s
    [("racket" "koka" "multicore-ocaml") "native"]
    [("scheme-monadic" "haskell-monadic") "[DPJS]"]
    [("scheme-oleg" "haskell-oleg") "[K]"]
    [else (string-titlecase s)]))

(define (factor s)
  (list (smaller "×") s))

(define (header s)
  (cond
    [(string? s) (italic s)]
    [(list? s) (map header s)]
    [else s]))

;; ----------------------------------------

(define (scheme-table)
  (define timess (make-hasheq))
  (call-with-input-file/maybe
   "bmtimes/scheme.txt"
   (lambda (in)
     (let loop ()
       (regexp-match #rx"^[ \r\n]*" in) ; discard whitespace
       (define c (peek-char in))
       (unless (eof-object? c)
         (case c
           [(#\[)
            (define r (read in))
            (define t (caaddr r))
            (when (and t (not (zero? t)))
              (define times (hash-ref timess (car r)
                                      (lambda ()
                                        (define ht (make-hasheq))
                                        (hash-set! timess (car r) ht)
                                        ht)))
              (hash-update! times (cadr r) (lambda (l) (cons t l)) null))
            (loop)]
           [else
            ;; other `make` output in transcript; discard it
            (regexp-match #rx"[^\r\n]*" in)
            (loop)])))))

  (define rowss
    (for/hasheq ([(k times) (in-hash timess)])
      (values k
              (for/hasheq ([(name l) (in-hash times)])
                (define median (list-ref (sort l <) (quotient (length l) 2)))
                (define avg (round (/ (apply + l) (length l))))
                (define stddev (if (= (length l) 1)
                                   0.0
                                   (sqrt (/ (for/sum ([t (in-list l)])
                                              (expt (- t avg) 2))
                                            (sub1 (length l))))))
                (values name (list median avg stddev))))))

  (define row-avg cadr)
  (define row-stdev caddr)
  (define N 1)

  (define racket-relevant? #f)
  (define P 0)

  (define (second-better? i1 i2 name)
    (define rows1 (hash-ref rowss i1))
    (define rows2 (hash-ref rowss i2))
    (define row1 (hash-ref rows1 name))
    (define row2 (hash-ref rows2 name))
    (and (< (row-avg row2)
            (- (row-avg row1) (max (* (/ P 100) (row-avg row1)) (* N (row-stdev row1)))))))

  (define (record-second-better i1 i2 #:attach-relevant? [attach-relevant? #f])
    (define rows1 (hash-ref rowss i1 #f))
    (when rows1
      (for ([name (in-hash-keys rows1)]
            #:when (second-better? i1 i2 name))
        (hash-set! relevant name #t)
        (when attach-relevant?
          (hash-set! attach-relevant name #t)))))

  (define (record-better i1 i2 #:attach-relevant? [attach-relevant? #f])
    (record-second-better i1 i2 #:attach-relevant? attach-relevant?)
    (record-second-better i2 i1 #:attach-relevant? attach-relevant?))

  (define attach-relevant (make-hasheq))
  (define relevant (make-hasheq))

  (record-better 'scheme-current 'scheme-attach #:attach-relevant? #t)
  (record-better 'scheme-attach 'scheme-racket)
  (record-better 'scheme-current 'scheme-racket #:attach-relevant? racket-relevant?)

  (define relevants (sort (hash-keys attach-relevant)
                          symbol<?))

  (define (->ratio n)
    (factor (~r #:precision '(= 2) n)))

  (define (->% n)
    (format "~a%" (inexact->exact (round (* n 100)))))

  (define n-irrelevant (- (hash-count (hash-ref rowss 'scheme-current #hasheq()))
                          (length relevants)))

  (tabular
   #:style 'block
   #:sep (hspace 2)
   #:column-properties '(right right center)
   (append 
    (list*
     (map header (list "" "unmod" "attach" "CS" "max"))
     (map header (list "benchmark" "time" "time" "time" "relstdev"))
     (for/list ([name (in-list relevants)])
       (define c (hash-ref (hash-ref rowss 'scheme-current) name))
       (define a (hash-ref (hash-ref rowss 'scheme-attach) name))
       (define r (hash-ref (hash-ref rowss 'scheme-racket) name))
       (list
        (tt (symbol->string name))
        (list (number->string (row-avg a)) " ms")
        (->ratio (/ (row-avg a) (row-avg c)))
        (->ratio (/ (row-avg r) (row-avg c)))
        (->% (max (/ (row-stdev c) (row-avg c))
                  (/ (row-stdev a) (row-avg a))
                  (/ (row-stdev r) (row-avg r)))))))
    (list
     (list
      (format "Not shown: ~a more with attach within ~a stddev"
              n-irrelevant
              N
              #;
              (if (= N 1) "" "s"))
      'cont 'cont 'cont 'cont)))))

;; ----------------------------------------

(define (ctc-table)
  (factor1-table (call-with-input-file/maybe
                  "bmtimes/ctc.txt"
                  (lambda (in)
                    (parse-times in list #:rel-stdev? #t)))
                 #:label "contract mode"
                 #:base "racketcs"
                 #:nexts '("racketcs-imitate")
                 #:base-label "buitin"
                 #:next-labels '("imitate")))

(define (opt-ctc-table)
  (factor1-table (call-with-input-file/maybe
                  "bmtimes/ctc.txt"
                  (lambda (in)
                    (parse-times in list #:rel-stdev? #t)))
                 #:label "contract mode"
                 #:base "racketcs"
                 #:nexts opts
                 #:base-label "Racket"
                 #:next-labels opt-next-labels
                 #:time-labels opt-time-labels
                 #:relstdev-label "stdev"))

(define (app-table)
  (define rows
    (call-with-input-file/maybe
     "bmtimes/app.txt"
     (lambda (in)
       (parse-times in list #:rel-stdev? #t))))

  (define (lookup group impl)
    (for/first ([row (in-list rows)]
                #:when (and (equal? group (car row))
                            (equal? impl (cadr row))))
      (list-tail row 2)))

  (define (lookup-delta group-a group-b impl)
    (define as (lookup group-a impl))
    (define bs (lookup group-b impl))
    (define (-s a b) (number->string (- (string->number a) (string->number b))))
    (define (+% a b) a)
    (list (-s (car as) (car bs))
          (-s (cadr as) (cadr bs))
          (+% (caddr as) (caddr bs))))

  (define imitate-name "imitate")
  (define builtin-name "builtin")

  (define alog-name "ActivityLog import")
  (define oliver-name "OL1V3R gauss.175.smt2")
  (define megaparsack-name "Megaparsack JSON")
  (define md-render-name "Markdown Reference")
  (define cish-name "Xsmith cish")

  (define tentative-new-rows
    (list
     (list* alog-name
            imitate-name
            (lookup "alog" "racketcs-imitate"))
     (list* cish-name
            imitate-name
            (lookup "cish" "racketcs-imitate"))
     (list* megaparsack-name
            imitate-name
            (lookup "megaparsack" "racketcs-imitate"))
     (list* md-render-name
            imitate-name
            (lookup "md-render" "racketcs-imitate"))
     (list* oliver-name
            imitate-name
            (lookup "OL1V3R" "racketcs-imitate"))

     (list* alog-name
            builtin-name
            (lookup "alog" "racketcs"))
     (list* cish-name
            builtin-name
            (lookup "cish" "racketcs"))
     (list* megaparsack-name
            builtin-name
            (lookup "megaparsack" "racketcs"))
     (list* md-render-name
            builtin-name
            (lookup "md-render" "racketcs"))
     (list* oliver-name
            builtin-name
            (lookup "OL1V3R" "racketcs"))))

  (define new-rows (for/list ([row (in-list tentative-new-rows)]
                              #:when (cddr row))
                     row))
  
  (factor1-table new-rows
                 #:label "application"
                 #:base "builtin"
                 #:nexts '("imitate")
                 #:base-label "builtin"
                 #:next-labels '("imitate")))

(define (opt-app-table)
  (factor1-table (call-with-input-file/maybe
                  "bmtimes/app.txt"
                  (lambda (in)
                    (parse-times in list #:rel-stdev? #t)))
                 #:label "contract mode"
                 #:base "racketcs"
                 #:nexts opts
                 #:base-label "Racket"
                 #:next-labels opt-next-labels
                 #:time-labels opt-time-labels
                 #:relstdev-label "stdev"
                 #:renames #hash(("alog" . "ActivityLog")
                                 ("cish" . "Xsmith")
                                 ("megaparsack" . "Megaparsack")
                                 ("md-render" . "Markdown"))))

(define (factor1-table rows
                       #:label label
                       #:base base-name
                       #:nexts next-names
                       #:base-label base-label
                       #:next-labels next-labels
                       #:time-labels [time-labels #f]
                       #:skip-bms [skip-bms '()]
                       #:relstdev-label [relstdev-label "relstdev"]
                       #:sep [sep (hspace 2)]
                       #:ratio-prefix [ratio-prefix ""]
                       #:renames [renames #hash()])
  (define (->n s) (string->number s))
  (define (%->n s) (/ (string->number (substring s 0 (sub1 (string-length s)))) 100.0))
  (define (fmt n) (list ratio-prefix (factor (~r #:precision '(= 2) n))))
  (define (fmt% n) (format "~a%" (inexact->exact (floor (* n 100)))))

  (define ratios
    (for/list ([row (in-list rows)]
               #:when (equal? base-name (cadr row))
               #:unless (member (car row) skip-bms))
      (define i-rows
        (for/list ([next-name (in-list next-names)])
          (or (for/first ([i-row (in-list rows)]
                          #:when (and (equal? next-name (cadr i-row))
                                      (equal? (car row) (car i-row))))
                i-row)
              (begin
                (log-error "warning: missing ~s ~s" next-name (car row))
                row))))
      (define avg (->n (list-ref row 3)))
      (define var (%->n (list-ref row 4)))
      (define i-avgs (for/list ([i-row (in-list i-rows)])
                       (->n (list-ref i-row 3))))
      (define i-vars (for/list ([i-row (in-list i-rows)])
                       (%->n (list-ref i-row 4))))

      (define ratios (for/list ([i-avg (in-list i-avgs)])
                       (if (zero? avg) 0.0 (/ i-avg avg))))
      
      (define name (let ([name (car row)])
                     (hash-ref renames name name)))

      (append
       (list (tt name "")
             (list (number->string avg) " ms"))
       (map fmt ratios)
       (list
        (list (fmt% (apply max var i-vars)))))))

  (tabular
   #:style 'block
   #:sep sep
   #:column-properties '(right)
   (append
    (list*
     (map header (append (list "" base-label) next-labels (list "max")))
     (map header (append (list label "time") (or time-labels (map (lambda (l) "time") next-labels)) (list relstdev-label)))
     ratios))))
